<?php

declare(strict_types=1);

namespace Drupal\track_usage\Hook;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\Entity\ConfigEntityStorageInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Hook\Attribute\Hook;
use Drupal\track_usage\Entity\TrackConfigInterface;
use Drupal\track_usage\Model\Operation;
use Drupal\track_usage\RecorderInterface;
use Drupal\track_usage\Trait\BackwardsCompatibilityTrait;
use Drupal\track_usage\Trait\EntityUtilityTrait;

/**
 * Records usage on insert and update.
 */
class RecordUsageHook {

  use BackwardsCompatibilityTrait;
  use EntityUtilityTrait;

  public function __construct(
    protected readonly ConfigFactoryInterface $configFactory,
    protected readonly RecorderInterface $recorder,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly Connection $db,
  ) {}

  /**
   * Records usages for entity.
   */
  #[Hook('entity_insert')]
  #[Hook('entity_update')]
  public function record(EntityInterface $entity): void {
    $storage = $this->entityTypeManager->getStorage('track_usage_config');
    assert($storage instanceof ConfigEntityStorageInterface);

    $ids = $storage->getQuery()
      ->condition('status', TRUE)
      ->condition('realTimeRecording', TRUE)
      ->execute();

    foreach ($storage->loadMultiple($ids) as $config) {
      assert($config instanceof TrackConfigInterface);
      if (!$config->isSource($entity) && !$config->isTraversable($entity) && !$config->isTarget($entity)) {
        // The entity means nothing for this configuration context.
        continue;
      }

      if ($config->isSource($entity) && $this->isRevisionable($entity) && $config->trackOnlyActiveRevision()) {
        // If the active revision has changed, and we don't track revisions,
        // remove the old active revision record.
        if ($this->getOriginal($entity) && $this->getOriginal($entity)->getRevisionId() != $entity->getRevisionId()) {
          $this->recorder->cleanup($this->getOriginal($entity), Operation::DeleteRevision);
        }
      }

      $this->recorder->record($entity, $config);
    }
  }

  /**
   * Removes usages for a deleted entity.
   */
  #[Hook('entity_delete')]
  public function onEntityDelete(EntityInterface $entity): void {
    $this->recorder->cleanup($entity, Operation::Delete);
  }

  /**
   * Removes usages for a deleted entity translation or language.
   */
  #[Hook('entity_translation_delete')]
  public function onEntityTranslationDelete(EntityInterface $entity): void {
    $this->recorder->cleanup($entity, Operation::DeleteTranslation);
  }

  /**
   * Removes usages for a deleted entity revision.
   */
  #[Hook('entity_revision_delete')]
  public function onEntityRevisionDelete(EntityInterface $entity): void {
    $this->recorder->cleanup($entity, Operation::DeleteRevision);
  }

  /**
   * Acts when a track usage config entity is updated.
   */
  #[Hook('track_usage_config_update')]
  public function onTrackConfigUpdate(TrackConfigInterface $config): void {
    $original = $this->getOriginal($config);
    assert($original instanceof TrackConfigInterface);

    // This config has been disabled. Cleanup the usage records.
    if (!$config->status() && $original->status()) {
      $this->recorder->cleanup($config, Operation::Delete);
    }

    // The configuration has changed. Update its records.
    if (serialize($config->toArray()) != serialize($original->toArray())) {
      $records = $this->db->select(RecorderInterface::TABLE)
        ->fields(RecorderInterface::TABLE, ['source_type', 'source_id_string', 'source_revision'])
        ->condition('config', $config->id())
        ->execute()
        ->fetchAll($this->fetchMode(\PDO::FETCH_NUM));

      // Remove all recorded usages for this config first.
      $this->recorder->cleanup($config, Operation::Delete);

      foreach ($records as $record) {
        $record[] = $config->id();
        $this->recorder->recordRaw(...$record);
      }
    }
  }

}
