<?php

declare(strict_types=1);

namespace Drupal\track_usage;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\Delete;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Entity\RevisionableStorageInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\track_usage\Entity\TrackConfigInterface;
use Drupal\track_usage\Model\EntityRole;
use Drupal\track_usage\Model\Usage;
use Drupal\track_usage\Trait\BackwardsCompatibilityTrait;
use Drupal\track_usage\Trait\EntityUtilityTrait;
use Drupal\track_usage\Model\Operation;
use Drupal\track_usage\Trait\IdColumnTrait;

/**
 * Default implementation of `track_usage.recorder` service.
 */
class Recorder implements RecorderInterface {

  use BackwardsCompatibilityTrait;
  use EntityUtilityTrait;
  use IdColumnTrait;

  /**
   * List of inserted/updated entities to be passed to destruct method.
   *
   * @var array<array-key, non-empty-string>
   */
  protected array $update = [];

  /**
   * List of deletions to perform record cleanup.
   *
   * @var array{\Drupal\Core\Entity\EntityInterface, Operation}
   */
  protected array $cleanup = [];

  public function __construct(
    protected readonly TrackerInterface $tracker,
    protected readonly Connection $db,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly EntityFieldManagerInterface $entityFieldManager,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function record(EntityInterface $entity, TrackConfigInterface $config): void {
    $this->recordRaw($entity->getEntityTypeId(), $entity->id(), $this->getRevisionId($entity), $config->id());
  }

  /**
   * {@inheritdoc}
   */
  public function recordRaw(string $entityTypeId, int|string $entityId, int|string $revisionId, string $configId): void {
    $this->update["$entityTypeId:$entityId:$revisionId"] = $configId;
  }

  /**
   * {@inheritdoc}
   */
  public function cleanup(EntityInterface $entity, Operation $operation): void {
    $this->cleanup[] = [$entity, $operation];
  }

  /**
   * {@inheritdoc}
   */
  public function destruct(): void {
    $this->optimizeStack();
    $this->performCleanup();

    while ($this->update) {
      $key = key($this->update);
      $configId = array_shift($this->update);

      if (!$entity = $this->getEntityInstance($key)) {
        continue;
      }
      if (!$config = $this->getConfig($configId)) {
        continue;
      }

      if ($config->isTraversable($entity)) {
        assert($entity instanceof FieldableEntityInterface);
        // Re-queue all source entities where this entity is involved.
        $this->requeueSourceEntities($entity, $config);
        continue;
      }

      if (!$config->isSource($entity)) {
        continue;
      }
      assert($entity instanceof FieldableEntityInterface);

      $entityTypeId = $entity->getEntityTypeId();
      $entityIdAsInteger = $this->getIntegerIdColumnValue($entity);
      $entityIdAsString = $this->getStringIdColumnValue($entity);

      foreach ($this->getTranslations($entity) as $translation) {
        foreach ($this->tracker->track($translation, $config) as $usage) {
          assert($usage instanceof Usage);
          $meta = json_encode((object) ['paths' => $usage->paths]);
          $this->db->merge(self::TABLE)->keys([
            'config' => $config->id(),
            'source_type' => $entityTypeId,
            'source_id_int' => $entityIdAsInteger,
            'source_id_string' => $entityIdAsString,
            'source_langcode' => $translation->language()->getId(),
            'source_revision' => $this->getRevisionId($translation),
            'target_type' => $usage->targetEntity->getEntityTypeId(),
            'target_id_int' => $this->getIntegerIdColumnValue($usage->targetEntity),
            'target_id_string' => $this->getStringIdColumnValue($usage->targetEntity),
          ])->fields(['meta' => $meta])->execute();
        }
      }
    }
  }

  /**
   * Optimizes the operation stacks.
   *
   * In some circumstances, in the same request, an entity might be scheduled
   * for update but also for deletion. This method is analyzing the update stack
   * and optimizes to avoid useless operations.
   */
  protected function optimizeStack(): void {
    foreach ($this->cleanup as [$entity, $operation]) {
      assert($entity instanceof EntityInterface);

      // All update operations belonging to a deleted config are obsolete.
      if ($operation === Operation::Delete && $entity->getEntityTypeId() === 'track_usage_config') {
        $this->update = array_filter($this->update, fn(string $configId): bool => $entity->id() !== $configId);
        continue;
      }

      // Remove any update operation concerning a deleted entity or revision.
      $this->update = array_filter(
        $this->update,
        function (string $key) use ($entity, $operation): bool {
          return match ($operation) {
            Operation::Delete => !str_starts_with($key, "{$entity->getEntityTypeId()}:{$entity->id()}:"),
            Operation::DeleteRevision => $key !== $this->getKey($entity),
            // Deleting a translation cannot collide with an update.
            Operation::DeleteTranslation => TRUE,
          };
        },
        ARRAY_FILTER_USE_KEY,
      );
    }
  }

  /**
   * Performs cleanup of records.
   */
  protected function performCleanup(): void {
    while ($this->cleanup) {
      [$entity, $operation] = array_shift($this->cleanup);
      match($operation) {
        Operation::Delete => $this->onEntityDelete($entity),
        Operation::DeleteRevision => $this->onEntityRevisionDelete($entity),
        Operation::DeleteTranslation => $this->onEntityTranslationDelete($entity),
      };

      // The deleted entity might be a traversable entity, part of usage paths.
      if ($entity instanceof FieldableEntityInterface) {
        $configs = $this->entityTypeManager->getStorage('track_usage_config')->loadMultiple();
        foreach ($configs as $config) {
          if ($config->isTraversable($entity)) {
            assert($entity instanceof FieldableEntityInterface);
            // Re-queue all source entities where this entity is involved.
            $this->requeueSourceEntities($entity, $config);
          }
        }
      }
    }
  }

  /**
   * Reacts after an entity has been being deleted.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being deleted.
   */
  protected function onEntityDelete(EntityInterface $entity): void {
    $entityTypeId = $entity->getEntityTypeId();

    // A track usage config was deleted. Remove all its records.
    if ($entityTypeId === 'track_usage_config') {
      $this->db->delete(self::TABLE)->condition('config', $entity->id())->execute();
      return;
    }
    // A language has been deleted.
    elseif ($entityTypeId === 'configurable_language') {
      $this->db->delete(self::TABLE)->condition('source_langcode', $entity->id())->execute();
      return;
    }

    $entityId = $this->getIdColumnValueByEntity($entity);

    $query = $this->db->delete(self::TABLE);
    $or = $query->orConditionGroup()
      ->condition(
        // Delete records when the entity acts like a target entity.
        $query->andConditionGroup()
          ->condition('target_type', $entityTypeId)
          ->condition($this->getIdColumnName($entityTypeId, EntityRole::Target), $entityId)
      )
      ->condition(
        // Or when the entity acts like a source entity.
        $query->andConditionGroup()
          ->condition('source_type', $entityTypeId)
          ->condition($this->getIdColumnName($entityTypeId, EntityRole::Source), $entityId)
      );
    $query->condition($or)->execute();
  }

  /**
   * Reacts after an entity revision has been being deleted.
   *
   * @param \Drupal\Core\Entity\EntityInterface $revision
   *   The entity revision being deleted.
   */
  public function onEntityRevisionDelete(EntityInterface $revision): void {
    $this->getQuery($revision)->condition('source_revision', $revision->getRevisionId())->execute();
  }

  /**
   * Reacts after an entity translation has been being deleted.
   *
   * @param \Drupal\Core\Entity\EntityInterface $translation
   *   The entity translation being deleted.
   */
  protected function onEntityTranslationDelete(EntityInterface $translation): void {
    $langcode = $translation->language()->getId();
    $this->getQuery($translation)->condition('source_langcode', $langcode)->execute();
  }

  /**
   * Provides a base query to delete records.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity revision or translation.
   *
   * @return \Drupal\Core\Database\Query\Delete
   *   A delete query.
   */
  protected function getQuery(EntityInterface $entity): Delete {
    $entityTypeId = $entity->getEntityTypeId();
    $entityId = $this->getIdColumnValueByEntity($entity);
    return $this->db->delete(self::TABLE)
      ->condition('source_type', $entityTypeId)
      ->condition($this->getIdColumnName($entityTypeId, EntityRole::Source), $entityId);
  }

  /**
   * Returns entity translations.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity.
   *
   * @return iterable<FieldableEntityInterface>
   *   Translated entities.
   */
  protected function getTranslations(FieldableEntityInterface $entity): iterable {
    if ($entity instanceof TranslatableInterface) {
      foreach ($entity->getTranslationLanguages() as $langcode => $language) {
        if ($entity->hasTranslation($langcode)) {
          yield $entity->getTranslation($langcode);
        }
      }
    }
    else {
      yield $entity;
    }
  }

  /**
   * Re-queues source entities if a traversable entity has changed.
   *
   * When a traversable entity is updated or deleted, it might impact the usage
   * of source entities. We search if this traversable entity is part of the
   * usage path and re-queue the source entities where this traversable entity
   * is involved.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $traversableEntity
   *   The affected traversable entity.
   * @param \Drupal\track_usage\Entity\TrackConfigInterface $config
   *   The track config.
   */
  protected function requeueSourceEntities(FieldableEntityInterface $traversableEntity, TrackConfigInterface $config): void {
    $results = $this->db->select(self::TABLE)
      ->distinct()
      ->fields(self::TABLE, ['config', 'source_type', 'source_id_string', 'source_id_int', 'source_revision'])
      ->where("JSON_CONTAINS(meta, JSON_QUOTE(:entity), '\$.paths')", [
        ':entity' => $this->getKey($traversableEntity),
      ])
      ->execute()
      ->fetchAll($this->fetchMode(\PDO::FETCH_NUM));

    foreach ($results as [$configId, $entityTypeId, $entityIdAsString, $entityIdAsInt, $revisionId]) {
      $key = "$entityTypeId:$entityIdAsString:$revisionId";
      if (!isset($this->update[$key])) {

        $storage = $this->entityTypeManager->getStorage($entityTypeId);
        $entity = $storage instanceof RevisionableStorageInterface && $revisionId > 0 ? $storage->loadRevision($revisionId) : $storage->load($entityIdAsString);

        // Re-queue this source entity.
        if ($entity) {
          $this->update[$key] = $configId;
        }

        $entityIdColumn = $this->getIdColumnName($entityTypeId, EntityRole::Source);
        $entityId = $this->getIdColumnValue($entityTypeId, $entityIdAsString);

        $this->db->delete(self::TABLE)
          ->condition('config', $config->id())
          ->condition('source_type', $entityTypeId)
          ->condition($entityIdColumn, $entityId)
          ->condition('source_revision', intval($revisionId))
          ->execute();
      }
    }
  }

  /**
   * Returns an entity instance out of a "key" string.
   *
   * @param string $entityAsString
   *   String same as the one returned by EntityUtilityTrait::getKey().
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The entity object or NULL if an entity cannot be determined.
   *
   * @see \Drupal\track_usage\Trait\EntityUtilityTrait::getKey()
   */
  protected function getEntityInstance(string $entityAsString): ?EntityInterface {
    [$entityTypeId, $entityId, $revisionId] = explode(':', $entityAsString, 3);
    $revisionId = intval($revisionId);

    if ($revisionId !== 0 && $this->isRevisionable($entityTypeId)) {
      $method = 'loadRevision';
      $id = $revisionId;
    }
    else {
      $method = 'load';
      $id = $entityId;
    }

    return $this->entityTypeManager->getStorage($entityTypeId)->{$method}($id);
  }

  /**
   * Returns a track usage config if a valid ID has been passed.
   *
   * @param string $configId
   *   The track usage config ID.
   *
   * @return \Drupal\track_usage\Entity\TrackConfigInterface|null
   *   The track usage config object or NULL.
   */
  protected function getConfig(string $configId): ?TrackConfigInterface {
    return $this->entityTypeManager->getStorage('track_usage_config')->load($configId);
  }

}
